"""MobileAlerts internet gataway."""

from typing import Any, Awaitable, Callable, Dict, List, Optional

import asyncio
import logging
import re
import socket
import struct
import time
from ipaddress import IPv4Address

import aiohttp
from multidict import CIMultiDictProxy
from yarl import URL

from .sensor import Sensor

_LOGGER = logging.getLogger(__name__)

# SensorHandler = Callable[[Any, Sensor], Awaitable[None]]
#

#: all communication with the gateways are broadcasts
BROADCAST_ADDR = "255.255.255.255"

#: UDP port used by the gateway for comunnications
PORT = 8003

# Commands which acceps gateway via UDP:
DISCOVER_GATEWAYS = 1
#: Find any available gateway in the local network
FIND_GATEWAY = 2
#: Find a single available gateway in the local network
GET_CONFIG = 3
#: Request the configuration of the gateway
SET_CONFIG = 4
#: Set a new configuration. Gateway takes a few seconds to do the update
REBOOT = 5
#: A reboot takes about 10s for the gateway to be back up again

ORIG_PROXY_BYTE1 = 0x19
#: 'Magic' byte #1 to mark preserved original proxy settings
ORIG_PROXY_BYTE2 = 0x74
#: 'Magic' byte #2 to mark preserved original proxy settings


class SensorHandler:
    """Abstract class of MobileAlerts senso's handler."""

    async def sensor_added(self, sensor: Sensor) -> None:
        pass

    async def sensor_updated(self, sensor: Sensor) -> None:
        pass


class Gateway:
    """Controls MobileAlerts internet gataway."""

    def __init__(
        self,
        gateway_id: str,
        local_ip_address: Optional[str] = None,
    ) -> None:
        self._id: bytes = bytes.fromhex(gateway_id)
        self._local_ip_address: Optional[str] = local_ip_address
        self._handler: Optional[SensorHandler] = None
        self._version = "1.50"
        self._last_seen: Optional[float] = None
        self._attached = False
        self._orig_use_proxy: Any = None
        self._orig_proxy: Any = None
        self._orig_proxy_port: Any = None
        self._dhcp_ip: Any = None
        self._use_dhcp: Any = None
        self._fixed_ip: Any = None
        self._fixed_netmask: Any = None
        self._fixed_gateway: Any = None
        self._name: Any = None
        self._server: Any = None
        self._use_proxy: Any = None
        self._proxy: Any = None
        self._proxy_port: Any = None
        self._fixed_dns: Any = None
        self._send_data_to_cloud = True
        self._sensors: Dict[str, Sensor] = dict()
        self._initialized = False
        self._is_online = False

    async def init(
        self,
        config: Optional[bytes] = None,
    ) -> bool:
        if config is None:
            config = await self.get_config()
        if config is not None:
            return self.parse_config(config)
        else:
            return False

    def _check_init(self) -> None:
        if not self._initialized:
            raise Exception("Gateway is not initialized")

    @staticmethod
    def prepare_socket(
        timeout: int,
        local_ip_address: Optional[str],
    ) -> socket.socket:
        """Prepares UDP socket to comunicate with the gateway."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.setblocking(False)
        sock.settimeout(timeout)
        if local_ip_address:
            sock.bind((local_ip_address, 0))
        else:
            sock.bind(("", 0))
        return sock

    @staticmethod
    def prepare_command(command: int, gateway_id: bytes) -> bytes:
        """Prepares command UDP packet to send."""
        packet = struct.pack(">H6sH", command, gateway_id, 10)
        return packet

    async def send_command(
        self, command: int, wait_for_result: bool = False, timeout: int = 2
    ) -> Optional[bytes]:
        """Sends command and optional data to the gateway."""
        packet = self.prepare_command(command, self._id)

        sock = self.prepare_socket(timeout, self._local_ip_address)
        try:
            sock.sendto(packet, (BROADCAST_ADDR, PORT))
            if wait_for_result:
                loop = asyncio.get_event_loop()
                config = await asyncio.wait_for(loop.sock_recv(sock, 256), timeout)
                self._last_seen = time.time()
                return config
            else:
                return None
        finally:
            sock.close()

    async def get_config(self, timeout: int = 30) -> Optional[bytes]:
        """Obtains configuration from the gateway."""
        _LOGGER.debug(
            "Gateway get_config for id=%s, timeout=%s", self.gateway_id, str(timeout)
        )
        result = None
        start_time = time.time()
        while True:
            if (time.time() - start_time) > timeout:
                break
            try:
                result = await self.send_command(FIND_GATEWAY, True, 5)
                break
            except socket.timeout:
                _LOGGER.debug("Gateway FIND_GATEWAY timeout")
                continue
            except asyncio.TimeoutError:
                _LOGGER.debug("Gateway FIND_GATEWAY timeout")
                continue

        self._is_online = result is not None
        return result

    @staticmethod
    def check_config(config: bytes) -> bool:
        return (
            config is not None
            and (len(config) >= 186)
            and (len(config) == int.from_bytes(config[8:10], "big"))
        )

    def parse_config(self, config: bytes) -> bool:
        """Parses configuration obtained from the gateway."""
        result = self.check_config(config) and (
            (self._id is None) or (self._id == config[2:8])
        )
        if result:
            orig_data = bytearray()
            self._id = config[2:8]
            self._dhcp_ip = IPv4Address(config[11:15])
            self._use_dhcp = config[15] != 0
            self._fixed_ip = IPv4Address(config[16:20])
            self._fixed_netmask = IPv4Address(config[20:24])
            self._fixed_gateway = IPv4Address(config[24:28])
            self._name = config[28 : config.find(0, 28, 49)].decode("utf-8")
            str_end_pos = config.find(0, 49, 114)
            if (
                config[str_end_pos + 1] == ORIG_PROXY_BYTE1
                and config[str_end_pos + 2] == ORIG_PROXY_BYTE2
            ):
                orig_data.extend(config[str_end_pos + 3 : 114])
            self._server = config[49:str_end_pos].decode("utf-8")
            self._use_proxy = config[114] != 0
            str_end_pos = config.find(0, 115, 180)
            self._proxy = config[115:str_end_pos].decode("utf-8")
            if (
                config[str_end_pos + 1] == ORIG_PROXY_BYTE1
                and config[str_end_pos + 2] == ORIG_PROXY_BYTE2
            ):
                orig_data.extend(config[str_end_pos + 3 : 180])
            self._proxy_port = int.from_bytes(config[180:182], "big")
            self._fixed_dns = IPv4Address(config[182:186])
            if len(orig_data) > 3:
                self._orig_use_proxy = orig_data[0]
                self._orig_proxy_port = int.from_bytes(orig_data[1:3], "big")
                str_end_pos = orig_data.find(0, 3)
                self._orig_proxy = orig_data[3:str_end_pos].decode("utf-8")
            self._last_seen = time.time()
            self._initialized = True
        return result

    async def update_config(self, timeout: int = 2) -> bool:
        """Updates configuration from the gateway."""
        config = await self.get_config(timeout)
        if config is not None:
            return self.parse_config(config)
        else:
            return False

    def set_config(self) -> None:
        """Set configuration to the gateway."""
        self._check_init()
        command = SET_CONFIG
        if self._orig_use_proxy is not None:
            orig_name_bytes = bytes(self._orig_proxy, "utf-8")
            orig_data_size = 3 + len(orig_name_bytes)
        else:
            orig_data_size = 0
        orig_data = bytearray(orig_data_size)
        if orig_data_size > 0:
            orig_data[0] = self._orig_use_proxy
            orig_data[1:3] = self._orig_proxy_port.to_bytes(2, "big")
            orig_data[3:orig_data_size] = orig_name_bytes
        orig_data_pos = 0
        packet_size = 181
        packet = bytearray(packet_size)
        packet[0:2] = command.to_bytes(2, "big")
        packet[2:8] = self._id
        packet[8:10] = packet_size.to_bytes(2, "big")
        packet[10] = self._use_dhcp
        packet[11:15] = self._fixed_ip.packed
        packet[15:19] = self._fixed_netmask.packed
        packet[19:23] = self._fixed_gateway.packed
        str_bytes = bytes(self._name, "utf-8")
        packet[23 : 23 + len(str_bytes)] = str_bytes
        str_bytes = bytes(21 - len(str_bytes))
        packet[44 - len(str_bytes) : 44] = str_bytes
        str_bytes = bytes(self._server, "utf-8")
        packet[44 : 44 + len(str_bytes)] = str_bytes
        str_bytes = bytearray(65 - len(str_bytes))
        if orig_data_pos < orig_data_size:
            str_bytes[1] = ORIG_PROXY_BYTE1
            str_bytes[2] = ORIG_PROXY_BYTE2
            orig_part_size = min(orig_data_size - orig_data_pos, len(str_bytes) - 3)
            str_bytes[3 : 3 + orig_part_size] = orig_data[
                orig_data_pos : orig_data_pos + orig_part_size
            ]
            orig_data_pos += orig_part_size
        packet[109 - len(str_bytes) : 109] = str_bytes
        packet[109] = self._use_proxy
        str_bytes = bytes(str(self._proxy), "utf-8")
        packet[110 : 110 + len(str_bytes)] = str_bytes
        str_bytes = bytearray(65 - len(str_bytes))
        if orig_data_pos < orig_data_size:
            str_bytes[1] = ORIG_PROXY_BYTE1
            str_bytes[2] = ORIG_PROXY_BYTE2
            orig_part_size = min(orig_data_size - orig_data_pos, len(str_bytes) - 3)
            str_bytes[3 : 3 + orig_part_size] = orig_data[
                orig_data_pos : orig_data_pos + orig_part_size
            ]
        packet[175 - len(str_bytes) : 175] = str_bytes
        packet[175:177] = self._proxy_port.to_bytes(2, "big")
        packet[177:181] = self._fixed_dns.packed

        sock = Gateway.prepare_socket(1, self._local_ip_address)
        try:
            sock.sendto(packet, (BROADCAST_ADDR, PORT))
        finally:
            sock.close()

    def reset_config(self) -> None:
        """Reset configuration of the gateway to default values."""
        self.name = "MOBILEALERTS-Gateway"
        self.use_dhcp = True
        self.fixed_ip = "192.168.1.222"
        self.fixed_netmask = "255.255.255.0"
        self.fixed_gateway = "192.168.1.254"
        self.fixed_dns = "192.168.1.253"
        self.server = "www.data199.com"
        self.use_proxy = False
        self.proxy = "192.168.1.1"
        self.proxy_port = 8080
        self.set_config()

    async def reboot(self, update_config: bool, timeout: int = 30) -> None:
        """Reboots the gateway and optional update configuration."""
        config = await self.send_command(REBOOT, update_config, timeout)
        if update_config and config is not None:
            self.parse_config(config)

    async def ping(self, reattach_to_proxy: bool, timeout: int = 30) -> bool:
        config = await self.get_config(timeout)
        if config:
            result: bool = self.check_config(config)
            if result:
                self._dhcp_ip = IPv4Address(config[11:15])
                self._use_dhcp = config[15] != 0
                self._fixed_ip = IPv4Address(config[16:20])
                if self._attached:
                    curr_use_proxy = config[114] != 0
                    str_end_pos = config.find(0, 115, 180)
                    curr_proxy = config[115:str_end_pos].decode("utf-8")
                    if (
                        (curr_use_proxy != self._use_proxy)
                        or (curr_proxy != self._proxy)
                    ) and reattach_to_proxy:
                        self.set_config()
            return result
        else:
            return False

    @staticmethod
    async def discover(
        local_ip_address: Optional[str] = None,
        timeout: int = 2,
    ) -> List["Gateway"]:
        """Broadcasts discover packet and yeld gateway objects created from resposes."""
        result = []
        discovered = []
        loop = asyncio.get_event_loop()

        sock = Gateway.prepare_socket(timeout, local_ip_address)
        packet = Gateway.prepare_command(DISCOVER_GATEWAYS, bytearray(6))

        try:
            sock.sendto(packet, (BROADCAST_ADDR, PORT))
            _LOGGER.debug("Gateways discovering packet sent")
            start_time = time.time()
            while True:
                try:
                    config = await asyncio.wait_for(loop.sock_recv(sock, 256), 1)
                    _LOGGER.debug("Gateways discovering response received %r", config)
                except socket.timeout:
                    break
                except asyncio.TimeoutError:
                    break
                if Gateway.check_config(config):
                    gateway_id = config[2:8]

                    if gateway_id in discovered:
                        continue
                    discovered.append(gateway_id)

                    gateway = Gateway(gateway_id.hex().upper(), local_ip_address)
                    await gateway.init(config)
                    result.append(gateway)
                if (time.time() - start_time) > timeout:
                    break
        finally:
            sock.close()

        return result

    def set_handler(
        self,
        handler: Optional[SensorHandler],
    ) -> None:
        self._handler = handler

    def attach_to_proxy(
        self,
        proxy: str,
        proxy_port: int,
        handler: Optional[SensorHandler],
    ) -> None:
        """Attachs the gateway to the proxy to read measuremnts.

        Existing proxy settings will be preserved
        """
        if self._orig_use_proxy is None:
            self._orig_use_proxy = self._use_proxy
            self._orig_proxy = self._proxy
            self._orig_proxy_port = self._proxy_port
        self._attached = True
        self._use_proxy = True
        self._proxy = IPv4Address(proxy)
        self._proxy_port = proxy_port
        self.set_handler(handler)
        self.set_config()

    def detach_from_proxy(self) -> None:
        """Detachs the gateway from the proxy and restore original settings."""
        if self._attached:
            self._use_proxy = self._orig_use_proxy
            self._proxy = self._orig_proxy
            self._proxy_port = self._orig_proxy_port
        self._attached = False
        self._orig_use_proxy = None
        self._orig_proxy = None
        self._orig_proxy_port = None
        self.set_handler(None)
        self.set_config()

    def handle_bootup_update(self, package: bytes) -> None:
        """Handle gateway's bootup update packet."""
        if (len(package) == 15) and (package[5:11] == self._id):
            _LOGGER.debug(
                "Gateway bootup timestamp %s",
                time.ctime(int.from_bytes(package[1:5], "big")),
            )
            self._version = (
                str(int.from_bytes(package[11:13], "big"))
                + "."
                + str(int.from_bytes(package[13:15], "big"))
            )
            self._last_seen = time.time()

    def add_sensor(self, sensor: Sensor) -> None:
        """Add sensor object."""
        self._sensors[sensor.sensor_id] = sensor

    @staticmethod
    async def get_sensor_name(sensor_id: str) -> Optional[str]:
        """Try to receive name of the sensor from the cloud."""
        try:
            url = (
                "https://measurements.mobile-alerts.eu/Home"
                + "/MeasurementDetails?deviceid=%s"
                + "&vendorid=9ac3a789-6f6a-47bf-8cf5-f076f532fe64"
                + "&appbundle=eu.mobile_alerts.mobilealerts"
            ) % (sensor_id)
            async with aiohttp.ClientSession() as session:
                async with session.get(str(url)) as response:
                    response_content: str = await response.text()
            match = re.search(r"<h3>(.*) [^ <]+<\/h3>", response_content)
            if match:
                name = match.group(1)
            else:
                name = None
            _LOGGER.debug("Discovered sensor name: %s", name)
            return name
        except Exception as e:
            _LOGGER.error("Error discovering sensor name: %r", e)
            return None

    async def create_sensor(self, sensor_id: str) -> Sensor:
        """Create new sensor object for given ID."""
        name: Optional[str] = await Gateway.get_sensor_name(sensor_id)
        result = Sensor(self, sensor_id, name)
        self.add_sensor(result)
        return result

    async def get_sensor(self, sensor_id: str) -> Sensor:
        """Return sensor object for given ID, creates the sensor if not exists."""
        result = self._sensors.get(sensor_id, None)
        if result is None:
            result = await self.create_sensor(sensor_id)
            _LOGGER.debug("New sensor is discovered: %r", result)
            if (
                result is not None
                and self._handler
                and callable(getattr(self._handler, "sensor_added", None))
            ):
                await self._handler.sensor_added(result)

        return result

    async def handle_sensor_update(self, package: bytes, package_checksum: int) -> None:
        """Handle update packet for one sensor."""
        _LOGGER.debug(
            "Update package %s, checksum %s",
            package.hex().upper(),
            hex(package_checksum),
        )

        checksum = 0
        for b in package:
            checksum += b
        checksum &= 0x7F

        if checksum == package_checksum:
            self._last_seen = time.time()
            sensor_id = package[6:12].hex().upper()
            sensor = await self.get_sensor(sensor_id)
            sensor.parse_packet(package)
            if self._handler and callable(
                getattr(self._handler, "sensor_updated", None)
            ):
                await self._handler.sensor_updated(sensor)
        else:
            _LOGGER.error(
                "Update package checksum error %s, checksum %s",
                package.hex().upper(),
                hex(package_checksum),
            )

    async def handle_sensors_update(self, packages: bytes) -> None:
        """Handle update packet for few sensors."""
        pos = 0
        packages_len = len(packages)

        while pos + 64 <= packages_len:
            await self.handle_sensor_update(
                packages[pos : pos + 63], packages[pos + 63]
            )
            pos += 64

    async def handle_update(
        self, code: str, packages: bytes, remote_ip: Optional[str] = None
    ) -> None:
        """Handle update packets."""
        _LOGGER.debug("Handling update from %s", remote_ip)
        self._is_online = True
        if self._use_dhcp and self._dhcp_ip != remote_ip:
            self._dhcp_ip = remote_ip
        if code == "00":
            self.handle_bootup_update(packages)
        elif code == "C0":
            await self.handle_sensors_update(packages)
        else:
            _LOGGER.error(
                "Unknnow update code %d, data %s",
                code,
                packages.hex().upper(),
            )

    async def resend_data_to_cloud(
        self,
        url: URL,
        headers: CIMultiDictProxy[str],
        content: bytes,
    ) -> None:
        """Resend gateway's PUT request to cloud server."""
        if self._send_data_to_cloud:
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.put(
                        str(url), headers=headers, data=content
                    ) as response:
                        response_content = await response.content.read()
                        _LOGGER.debug(
                            "Cloud response status: %s content: %s",
                            response.status,
                            response_content.hex().upper(),
                        )
            except Exception as e:
                _LOGGER.error("Error resending request to cloud: %r", e)

    @property
    def gateway_id(self) -> str:
        return self._id.hex().upper()

    @property
    def serial(self) -> str:
        return "80%s" % self._id[3:6].hex().upper()

    @property
    def version(self) -> str:
        return self._version

    @property
    def is_online(self) -> bool:
        return self._is_online

    @property
    def ip_address(self) -> Optional[str]:
        if self._is_online:
            return str(self._dhcp_ip) if self._use_dhcp else str(self._fixed_ip)
        else:
            return None

    @property
    def url(self) -> Optional[str]:
        if self._is_online:
            return "http://%s/" % self.ip_address
        else:
            return None

    @property
    def last_seen(self) -> Optional[float]:
        return self._last_seen

    @property
    def attached(self) -> bool:
        return self._attached

    @property
    def send_data_to_cloud(self) -> bool:
        return self._send_data_to_cloud

    @send_data_to_cloud.setter
    def send_data_to_cloud(self, value: bool) -> None:
        self._send_data_to_cloud = value

    @property
    def dhcp_ip(self) -> str:
        return str(self._dhcp_ip)

    @property
    def use_dhcp(self) -> bool:
        return bool(self._use_dhcp)

    @use_dhcp.setter
    def use_dhcp(self, value: bool) -> None:
        self._use_dhcp = value

    @property
    def fixed_ip(self) -> str:
        return str(self._fixed_ip)

    @fixed_ip.setter
    def fixed_ip(self, value: str) -> None:
        self._fixed_ip = IPv4Address(value)

    @property
    def fixed_netmask(self) -> str:
        return str(self._fixed_netmask)

    @fixed_netmask.setter
    def fixed_netmask(self, value: str) -> None:
        self._fixed_netmask = IPv4Address(value)

    @property
    def fixed_gateway(self) -> str:
        return str(self._fixed_gateway)

    @fixed_gateway.setter
    def fixed_gateway(self, value: str) -> None:
        self._fixed_gateway = IPv4Address(value)

    @property
    def name(self) -> str:
        return str(self._name)

    @name.setter
    def name(self, value: str) -> None:
        if len(bytes(value, "utf-8")) > 20:
            raise ValueError("Name is too long")
        self._name = value

    @property
    def server(self) -> str:
        return str(self._server)

    @server.setter
    def server(self, value: str) -> None:
        if len(bytes(value, "utf-8")) > 64:
            raise ValueError("Server address is too long")
        self._server = value

    @property
    def use_proxy(self) -> bool:
        return bool(self._use_proxy)

    @use_proxy.setter
    def use_proxy(self, value: bool) -> None:
        self._use_proxy = value

    @property
    def proxy(self) -> str:
        return str(self._proxy)

    @proxy.setter
    def proxy(self, value: str) -> None:
        if len(bytes(value, "utf-8")) > 64:
            raise ValueError("Proxy server address is too long")
        self._proxy = value

    @property
    def proxy_port(self) -> int:
        return int(self._proxy_port)

    @proxy_port.setter
    def proxy_port(self, value: int) -> None:
        if value < 0 or value >= 64 * 1024:
            raise ValueError("Invalid proxy port number")
        self._proxy_port = value

    @property
    def fixed_dns(self) -> str:
        return str(self._fixed_dns)

    @fixed_dns.setter
    def fixed_dns(self, value: str) -> None:
        self._fixed_dns = IPv4Address(value)

    @property
    def orig_use_proxy(self) -> bool:
        return bool(self._orig_use_proxy)

    @property
    def orig_proxy(self) -> str:
        return str(self._orig_proxy)

    @property
    def orig_proxy_port(self) -> int:
        return int(self._orig_proxy_port)

    @property
    def sensors(self) -> List[Sensor]:
        return [*self._sensors.values()]

    def __repr__(self) -> str:
        """Return a formal representation of the gateway."""
        return (
            "%s.%s(%s(%s), "
            "gateway_id=%s, "
            "version=%r, "
            "last_seen=%r, "
            "attached=%r, "
            "send_data_to_cloud=%r, "
            "dhcp_ip=%r, "
            "use_dhcp=%r, "
            "fixed_ip=%r, "
            "fixed_netmask=%r, "
            "fixed_gateway=%r, "
            "fixed_dns=%r, "
            "server=%r, "
            "use_proxy=%r, "
            "proxy=%r, "
            "proxy_port=%r, "
            "orig_use_proxy=%r, "
            "orig_proxy=%r, "
            "orig_proxy_port=%r"
            ")"
        ) % (
            self.__class__.__module__,
            self.__class__.__qualname__,
            self.name,
            self.serial,
            self.gateway_id,
            self.version,
            time.ctime(self.last_seen) if self.last_seen is not None else "never",
            self.attached,
            self.send_data_to_cloud,
            self.dhcp_ip,
            self.use_dhcp,
            self.fixed_ip,
            self.fixed_netmask,
            self.fixed_gateway,
            self.fixed_dns,
            self.server,
            self.use_proxy,
            self.proxy,
            self.proxy_port,
            self.orig_use_proxy,
            self.orig_proxy,
            self.orig_proxy_port,
        )

    def __str__(self) -> str:
        """Return a readable representation of the gateway."""
        return (
            "%s V%s, SerialNo: %s (id: %s)\n"
            "Use DHCP: %s\n"
            "DHCP IP: %s\n"
            "Fixed IP: %s\n"
            "Fixed Netmask: %s\n"
            "Fixed Gateway: %s\n"
            "Fixed DNS: %s\n"
            "Cloud Server: %s\n"
            "Use Proxy: %s\n"
            "Proxy Server: %s\n"
            "Proxy Port: %s\n"
            "Send data to cloud: %s\n"
            "Last Contact: %s"
        ) % (
            self.name,
            self.version,
            self.serial,
            self.gateway_id,
            "Yes" if self.use_dhcp else "No",
            self.dhcp_ip,
            self.fixed_ip,
            self.fixed_netmask,
            self.fixed_gateway,
            self.fixed_dns,
            self.server,
            "Yes" if self.use_proxy else "No",
            self.proxy,
            self.proxy_port,
            "Yes" if self.send_data_to_cloud else "No",
            time.ctime(self.last_seen) if self.last_seen is not None else "never",
        )



"""MobileAlerts sensors."""

from typing import Any, Dict, List, Optional, Tuple, Union

import datetime
import logging
import time
from enum import IntEnum, auto

_LOGGER = logging.getLogger(__name__)


UNKNOWN = "Unknown"


class MeasurementType(IntEnum):
    """Types of measurments."""

    TEMPERATURE = auto()
    HUMIDITY = auto()
    WETNESS = auto()
    CO2 = auto()
    AIR_PRESSURE = auto()
    RAIN = auto()
    TIME_SPAN = auto()
    ALARM = auto()
    WIND_SPEED = auto()
    GUST = auto()
    WIND_DIRECTION = auto()
    DOOR_WINDOW = auto()
    KEY_PRESSED = auto()
    KEY_PRESS_TYPE = auto()


MEASUREMENT_TYPES: List[str] = [
    "Temperature",
    "Humidity",
    "Wetness",
    "CO2",
    "Air pressure",
    "Rain",
    "Time span",
    "Alarm",
    "Wind speed",
    "Gust",
    "Wind direction",
    "Door/Window",
    "Key pressed",
    "Key press type",
]


class MeasurementError(IntEnum):
    """Measurment errors."""

    ERROR = auto()
    OVERFLOW = auto()
    NOT_CALCULATED = auto()


MEASUREMENT_ERRORS: List[str] = [
    "error",
    "overflow",
    "not calculcated",
]


"""
Dictionary with information about sensors by ID: 
model of sensor, description (also used as default name) and
measurements update interval (in seconds)
"""
SENSOR_INFOS: Dict[int, Tuple[str, str, int]] = {
    0x01: (
        "MA10120",
        "Temperature sensor in/cable",
        7 * 60,  # 7 min
    ),
    0x02: (
        "MA10100",
        "Temperature sensor",
        7 * 60,  # 7 min
    ),
    0x03: (
        "MA10250",
        "Temperature/Humidity",
        7 * 60,  # 7 min
    ),
    0x04: (
        "MA10350",
        "Temperature/Humidity/Water Detector sensor",
        7 * 60,  # 7 min
    ),
    0x05: (
        "WL2000",
        "Air Quality Monitor",
        7 * 60,  # 7 min
    ),
    0x06: (
        "MA10700",
        "Temperature/Humidity/Pool sensor",
        7 * 60,  # 7 min
    ),
    0x07: (
        "MA10410",
        "Weather Station",
        6 * 60,  # 6 min
    ),
    0x08: (
        "MA10650",
        "Rain sensor",
        2 * 60 * 60,  # 2 hours
    ),
    0x09: (
        "MA10320",
        "Pro Temperature/Humidity/Ext.Temperature sensor",
        3 * 60 + 30,  # 3.5 min
    ),
    0x0A: (
        "MA10860",
        "Sensor for acoustical observation of detectors",
        0,  # Unknown
    ),
    0x0B: (
        "MA10660",
        "Wind sensor",
        # 2 * 60 * 60,  # 2 hours
        7 * 60,  # 7 min
    ),
    0x0E: (
        "TFA30.3312.02",
        "Professional Thermo/Hygro sensor",
        3 * 60 + 30,  # 3.5 min
    ),
    0x0F: (
        "MA10450",
        "Weather Station",
        7 * 60,  # 7 min
    ),
    0x10: (
        "MA10800",
        "Door/Window sensor",
        0,  # Unknown
    ),
    0x11: (
        "TFA30.3060.01",
        "4 Thermo-hygro-sensors",
        7 * 60,  # 7 min
    ),
    0x12: (
        "MA10230",
        "Humidity Guard",
        0,
    ),
    0x15: (
        "MA10880",
        "4 button switch",
        0,  # Unknown
    ),
    0x18: (
        "MA10238",
        "Air pressure monitor",
        6 * 60,  # 6 min
    ),
}


def _parse_temperature(
    value: bytes,
    attr: bool,
) -> Union[float, MeasurementError]:
    if len(value) != 2:
        raise ValueError("Invalid temperature value")
    result: int = int.from_bytes(value, "big")
    if attr:
        if result & (1 << 12) != 0:
            return MeasurementError.ERROR
        if result & (1 << 13) != 0:
            return MeasurementError.OVERFLOW
    negative = result & (1 << 10) != 0
    result &= 0x3FF
    if negative:
        result -= 1024
    return result * 0.1


def _parse_humidity(
    value: int,
    average: bool = False,
) -> Union[float, MeasurementError]:
    if average and value & 0x80 != 0:
        return MeasurementError.NOT_CALCULATED
    return value & 0x7F


def _parse_humidity_hr(
    value: bytes,
) -> float:
    if len(value) != 2:
        raise ValueError("Invalid air humidity value")
    result: int = int.from_bytes(value, "big")
    return (result & 0x1FF) / 10


def _parse_air_preasure(
    value: bytes,
) -> float:
    if len(value) != 2:
        raise ValueError("Invalid air preasure value")
    result: int = int.from_bytes(value, "big")
    return result / 10


def _parse_rain_time_span(value: int) -> int:
    time_unit = (value & 0xC000) >> 14
    if time_unit == 1:  # hours
        time_mult = 60 * 60
    elif time_unit == 2:  # minutes
        time_mult = 60
    else:  # seconds
        time_mult = 1
    return (value & 0x3FFF) * time_mult


def _parse_wind_direction(value: int) -> float:
    return ((value & 0xF0) >> 4) * 22.5


def _parse_wind_speed(
    value: int,
    hibit: int,
    himask: int,
) -> float:
    return ((value & 0xFF) | (0x100 if (hibit & himask != 0) else 0)) / 10


def _parse_wind_time_span(value: int) -> int:
    return (value & 0xFF) * 2


def _parse_door_window_time_span(value: int) -> int:
    time_unit = (value & 0x6000) >> 13
    if time_unit == 1:  # hours
        time_mult = 60 * 60
    elif time_unit == 2:  # minutes
        time_mult = 60
    else:  # seconds
        time_mult = 1
    return (value & 0x1FFF) * time_mult


class Measurement:
    """Hold value of one measurment done by the sensor."""

    def __init__(
        self,
        parent: "Sensor",
        type: MeasurementType,
        prefix: str = "",
        index: int = 0,
    ) -> None:
        """Create a new instance of the Measurement."""
        self._parent = parent
        self._type = type
        self._prefix = prefix
        self._index = index
        self._value: Any = None
        self._prior: Any = None

    @property
    def parent(self) -> "Sensor":
        return self._parent

    @property
    def type(self) -> MeasurementType:
        return self._type

    @property
    def name(self) -> str:
        name: str = MEASUREMENT_TYPES[int(self._type) - 1]
        return ((self._prefix + " " + name.lower()) if self._prefix else name) + (
            (" " + str(self._index)) if self._index > 0 else ""
        )

    @property
    def value(self):
        return self._value

    @property
    def index(self) -> int:
        return self._index

    @property
    def prefix(self) -> Optional[str]:
        return self._prefix

    @property
    def has_prior_value(self) -> bool:
        return self._prior is not None

    @property
    def prior_value(self):
        return self._prior

    def __repr__(self) -> str:
        """Return a formal representation of the measurement."""
        return (
            "%s.%s, "
            "type=%s, "
            'prefix="%s", '
            "index=%s, "
            "value=%r, "
            "prior_value=%r)"
        ) % (
            self.__class__.__module__,
            self.__class__.__qualname__,
            self._type,
            self._prefix,
            self._index,
            self._value,
            self._prior,
        )

    def unit(self) -> Union[str, List[str]]:
        if self._type == MeasurementType.TEMPERATURE:
            return "°C"
        elif self._type == MeasurementType.HUMIDITY:
            return "%"
        elif self._type == MeasurementType.WETNESS:
            return ["dry", "wet"]
        elif self._type == MeasurementType.CO2:
            return "ppm"
        elif self._type == MeasurementType.AIR_PRESSURE:
            return "hPa"
        elif self._type == MeasurementType.RAIN:
            return "mm"
        elif self._type == MeasurementType.TIME_SPAN:
            return "s"
        elif self._type == MeasurementType.ALARM:
            return ["calm", "alarm"]
        elif (
            self._type == MeasurementType.WIND_SPEED
            or self._type == MeasurementType.GUST
        ):
            return "m/s"
        elif self._type == MeasurementType.WIND_DIRECTION:
            return "°"
        elif self._type == MeasurementType.DOOR_WINDOW:
            return ["closed", "opened"]
        elif self._type == MeasurementType.KEY_PRESSED:
            return ["none", "green", "orange", "red", "yellow"]
        elif self._type == MeasurementType.KEY_PRESS_TYPE:
            return ["none", "short", "double", "long"]
        else:
            return ""

    def _value_to_str(self, value: Any) -> str:
        if value is None:
            return "unknown"
        elif type(value) is MeasurementError:
            return MEASUREMENT_ERRORS[int(value) - 1]
        elif self._type in [
            MeasurementType.TEMPERATURE,
            MeasurementType.HUMIDITY,
            MeasurementType.WIND_DIRECTION,
        ]:
            return str(round(value, 1)) + str(self.unit())
        elif self._type in [
            MeasurementType.RAIN,
            MeasurementType.AIR_PRESSURE,
            MeasurementType.CO2,
            MeasurementType.WIND_SPEED,
            MeasurementType.GUST,
        ]:
            return str(round(value, 1)) + " " + str(self.unit())
        elif self._type == MeasurementType.TIME_SPAN:
            return str(datetime.timedelta(seconds=value))
        elif self._type in [
            MeasurementType.WETNESS,
            MeasurementType.ALARM,
            MeasurementType.DOOR_WINDOW,
            MeasurementType.KEY_PRESSED,
            MeasurementType.KEY_PRESS_TYPE,
        ]:
            return self.unit()[int(value)]
        else:
            return "unknown"

    @property
    def value_str(self) -> str:
        return self._value_to_str(self._value)

    @property
    def prior_value_str(self) -> str:
        if self._prior is not None and hasattr(self._prior, "__len__"):
            result: str = "["
            i: int = 0
            l: int = len(self._prior)
            while True:
                result += self._value_to_str(self._prior[i])
                i += 1
                if i < l:
                    result += "; "
                else:
                    break
            return result + "]"
        else:
            return self._value_to_str(self._prior)

    def __str__(self) -> str:
        """Return a readable representation of the measurement."""
        result: str = ("%s: %s") % (self.name, self.value_str)
        if self.has_prior_value:
            result += ("; previous: %s") % (self.prior_value_str)
        return result

    def _set_temperature(
        self,
        value: bytes,
        prior: Optional[bytes] = None,
        attr: bool = True,
    ) -> None:
        self._value = _parse_temperature(value, attr)
        if prior is None:
            self._prior = None
        else:
            self._prior = _parse_temperature(prior, attr)

    def _add_prior_temperature(self, value: bytes) -> None:
        if self._prior is None:
            self._prior = []
        else:
            self._prior = [self._prior]
        self._prior.append(_parse_temperature(value, True))

    def _set_humidity(
        self,
        value: int,
        prior: Optional[int] = None,
        average: bool = False,
    ) -> None:
        self._value = _parse_humidity(value, average)
        if prior is None:
            self._prior = None
        else:
            self._prior = _parse_humidity(prior, average)

    def _set_humidity_hr(
        self,
        value: bytes,
        prior1: bytes,
        prior2: bytes,
    ) -> None:
        self._value = _parse_humidity_hr(value)
        self._prior = [_parse_humidity_hr(prior1), _parse_humidity_hr(prior2)]

    def _set_wetness(
        self,
        value: int,
    ) -> None:
        self._value = ((value & 0x02) != 0) or ((value & 0x01) == 0)

    def _set_CO2(
        self,
        value: bytes,
    ) -> None:
        if len(value) != 2:
            raise ValueError("Invalid air quality value")
        result: int = int.from_bytes(value, "big")
        if result & 0x100 != 0:
            self._value = MeasurementError.OVERFLOW
        else:
            self._value = (result & 0xFF) * 50

    def _set_air_preasure(
        self,
        value: bytes,
        prior: Optional[bytes] = None,
    ) -> None:
        self._value = _parse_air_preasure(value)
        if prior is None:
            self._prior = None
        else:
            self._prior = _parse_air_preasure(prior)

    def _set_rain(
        self,
        value: bytes,
    ) -> None:
        if len(value) != 2:
            raise ValueError("Invalid rain level value")
        result: int = int.from_bytes(value, "big")
        self._value = result * 0.25

    def _set_rain_time_span(
        self,
        values: bytes,
    ) -> None:
        if len(values) < 2:
            raise ValueError("Invalid timespan value")
        value = int.from_bytes(values[0:2], "big")
        self._value = _parse_rain_time_span(value)
        self._prior = None
        if len(values) > 2:
            index = 4
            while index < len(values):
                value = int.from_bytes(values[index - 2 : index], "big")
                if value == 0:
                    break
                value = _parse_rain_time_span(value)
                if value == 0:
                    break
                if self._prior is None:
                    self._prior = []
                self._prior.append(value)
                index += 2

    def _set_boolean(
        self,
        value: bytes,
        bitmask: int,
    ) -> None:
        if len(value) != 2:
            raise ValueError("Invalid value")
        result = int.from_bytes(value, "big")
        self._value = result & bitmask != 0

    def _add_wind_direction(
        self,
        index: int,
        value: int,
    ) -> None:
        if index == 0:
            self._value = _parse_wind_direction(value)
            self._prior = []
        else:
            self._prior.append(_parse_wind_direction(value))

    def _add_wind_speed(
        self,
        index: int,
        value: int,
        hibit: int,
        himask: int,
    ) -> None:
        if index == 0:
            self._value = _parse_wind_speed(value, hibit, himask)
            self._prior = []
        else:
            self._prior.append(_parse_wind_speed(value, hibit, himask))

    def _add_wind_time_span(
        self,
        index: int,
        value: int,
    ) -> None:
        if index == 0:
            self._value = _parse_wind_time_span(value)
            self._prior = []
        else:
            self._prior.append(_parse_wind_time_span(value))

    def _set_door_window_time_span(
        self,
        values: bytes,
    ) -> None:
        if len(values) < 2:
            raise ValueError("Invalid timespan value")
        value = int.from_bytes(values[0:2], "big")
        self._value = _parse_door_window_time_span(value)
        self._prior = None
        if len(values) > 2:
            index = 4
            while index < len(values):
                value = int.from_bytes(values[index - 2 : index], "big")
                if value == 0:
                    break
                value = _parse_door_window_time_span(value)
                if value == 0:
                    break
                if self._prior is None:
                    self._prior = []
                self._prior.append(value)
                index += 2

    def _set_key_pressed(
        self,
        value: int,
    ) -> None:
        self._value = (value & 0xF0) >> 4

    def _set_key_press_type(
        self,
        value: int,
    ) -> None:
        self._value = value & 0xF


class Sensor:
    """Receive data from Mobile Alerts/WeatherHub sensor."""

    def __init__(self, parent: Any, sensor_id: str, name: Optional[str] = None) -> None:
        self._id = sensor_id
        self._type_id = int(sensor_id[0:2], 16)

        self._parent = parent

        self._model, default_name, self._update_period = SENSOR_INFOS.get(
            self._type_id, (UNKNOWN, UNKNOWN, 0)
        )

        if name is None:
            name = f"{default_name} ({sensor_id})"
        self._name = name

        self._counter = -1
        self._low_battery = False
        self._by_event = False
        self._timestamp = time.time()
        self._last_update: Optional[bytes] = None

        self._name = name

        self._measurements: List[Measurement] = []
        self._three_byte_counter = False

        if self._type_id == 0x01 or self._type_id == 0x0F:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.TEMPERATURE, "Cable")
        elif self._type_id == 0x02:
            self._append(MeasurementType.TEMPERATURE)
        elif self._type_id == 0x03 or self._type_id == 0x0E:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.HUMIDITY)
        elif self._type_id == 0x04:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.HUMIDITY)
            self._append(MeasurementType.WETNESS)
        elif self._type_id == 0x05:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.HUMIDITY)
            self._append(MeasurementType.CO2)
            self._append(MeasurementType.TEMPERATURE, "Outdoor")
        elif self._type_id == 0x06:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.HUMIDITY)
            self._append(MeasurementType.TEMPERATURE, "Pool")
        elif self._type_id == 0x07:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.HUMIDITY)
            self._append(MeasurementType.TEMPERATURE, "Outdoor")
            self._append(MeasurementType.HUMIDITY, "Outdoor")
        elif self._type_id == 0x08:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.RAIN)
            self._append(MeasurementType.TIME_SPAN)
        elif self._type_id == 0x09:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.TEMPERATURE, "External")
            self._append(MeasurementType.HUMIDITY)
        elif self._type_id == 0x0A:
            self._append(MeasurementType.ALARM, "", 1)
            self._append(MeasurementType.ALARM, "", 2)
            self._append(MeasurementType.ALARM, "", 3)
            self._append(MeasurementType.ALARM, "", 4)
            self._append(MeasurementType.TEMPERATURE)
        elif self._type_id == 0x0B:
            self._three_byte_counter = True
            self._append(MeasurementType.WIND_DIRECTION)
            self._append(MeasurementType.WIND_SPEED)
            self._append(MeasurementType.GUST)
            self._append(MeasurementType.TIME_SPAN)
        elif self._type_id == 0x10:
            self._append(MeasurementType.DOOR_WINDOW)
            self._append(MeasurementType.TIME_SPAN)
        elif self._type_id == 0x11:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.HUMIDITY)
            self._append(MeasurementType.TEMPERATURE, "External", 1)
            self._append(MeasurementType.HUMIDITY, "External", 1)
            self._append(MeasurementType.TEMPERATURE, "External", 2)
            self._append(MeasurementType.HUMIDITY, "External", 2)
            self._append(MeasurementType.TEMPERATURE, "External", 3)
            self._append(MeasurementType.HUMIDITY, "External", 3)
        elif self._type_id == 0x12:
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.HUMIDITY)
            self._append(MeasurementType.HUMIDITY, "3h average")
            self._append(MeasurementType.HUMIDITY, "24h average")
            self._append(MeasurementType.HUMIDITY, "7d average")
            self._append(MeasurementType.HUMIDITY, "30d average")
        elif self._type_id == 0x15:
            self._append(MeasurementType.KEY_PRESSED)
            self._append(MeasurementType.KEY_PRESS_TYPE)
        elif self._type_id == 0x18:
            self._three_byte_counter = True
            self._append(MeasurementType.TEMPERATURE)
            self._append(MeasurementType.HUMIDITY)
            self._append(MeasurementType.AIR_PRESSURE)
        else:
            _LOGGER.error("Unknow sensor type: %s", hex(self._type_id))

    def __len__(self) -> int:
        return len(self._measurements)

    def __getitem__(self, key: int) -> Measurement:
        return self._measurements[key]

    def __repr__(self) -> str:
        """Return a formal representation of the sensor."""
        result: str = (
            "%s.%s(%s), "
            "id = %s, "
            "counter = %s, "
            "low_battery = %s, "
            "by_event = %s, "
            "timestamp = %s,"
            "last_update = %s"
        ) % (
            self.__class__.__module__,
            self.__class__.__qualname__,
            self._name,
            self._id,
            self._counter,
            self._low_battery,
            self._by_event,
            time.ctime(self._timestamp),
            self._last_update.hex().upper() if self._last_update is not None else None,
        )
        first = True
        for measurement in self._measurements:
            result += (", measurements: %r" if first else ", %r") % (measurement)
            first = False
        return result

    def __str__(self) -> str:
        """Return a readable representation of the sensor."""
        result: str = ("%s (id: %s, battery %s, last %s: %s)") % (
            self._name,
            self._id,
            "low" if self._low_battery else "good",
            "event" if self._by_event else "seen",
            time.ctime(self._timestamp),
        )
        for measurement in self._measurements:
            result += "\n" + str(measurement)
        return result

    def str_utc(self) -> str:
        """Return a readable representation of the sensor. Timestamp is formatted as UTC datetime"""
        timestamp_struct: time.struct_time = time.gmtime(self._timestamp)

        result: str = ("id: %s (battery %s, last %s: %s)") % (
            self._id,
            "low" if self._low_battery else "good",
            "event" if self._by_event else "seen",
            time.strftime("%Y-%m-%d %H:%M:%S", timestamp_struct),
        )
        for measurement in self._measurements:
            result += "\n" + str(measurement)
        return result

    def _append(
        self,
        type: MeasurementType,
        prefix: str = "",
        index: int = 0,
    ) -> None:
        self._measurements.append(Measurement(self, type, prefix, index))

    def _parse_packet_header(self, packet: bytes) -> bool:
        self._timestamp = int.from_bytes(packet[1:5], "big")
        if self._three_byte_counter:
            counter = int.from_bytes(packet[12:15], "big")
            self._low_battery = (counter & 0x800000) != 0
            self._by_event = (counter & 0x400000) != 0
            counter &= 0x3FFFFF
        else:
            counter = int.from_bytes(packet[12:14], "big")
            self._low_battery = (counter & 0x8000) != 0
            self._by_event = (counter & 0x4000) != 0
            counter &= 0x3FFF

        result = self._counter != counter
        if result:
            self._counter = counter

        return result

    def parse_packet(self, packet: bytes) -> None:
        if not self._parse_packet_header(packet):
            return
        if self._type_id == 0x01 or self._type_id == 0x0F:
            self[0]._set_temperature(packet[14:16], packet[18:20])
            self[1]._set_temperature(packet[16:18], packet[20:22])
        elif self._type_id == 0x02:
            self[0]._set_temperature(packet[14:16], packet[16:18])
        elif self._type_id == 0x03:
            self[0]._set_temperature(packet[14:16], packet[18:20])
            self[1]._set_humidity(packet[17], packet[21])
        elif self._type_id == 0x04:
            self[0]._set_temperature(packet[14:16], packet[19:21])
            self[1]._set_humidity(packet[17], packet[23])
            self[2]._set_wetness(packet[18])
        elif self._type_id == 0x05:
            self[0]._set_temperature(packet[16:18], packet[24:26])
            self[1]._set_humidity(packet[19], packet[27])
            self[2]._set_CO2(packet[20:22])
            self[3]._set_temperature(packet[14:16], packet[22:24])
        elif self._type_id == 0x06:
            self[0]._set_temperature(packet[14:16], packet[20:22])
            self[1]._set_humidity(packet[19], packet[25])
            self[2]._set_temperature(packet[16:18], packet[22:24])
        elif self._type_id == 0x07:
            self[0]._set_temperature(packet[14:16], packet[22:24])
            self[1]._set_humidity(packet[17], packet[25])
            self[2]._set_temperature(packet[18:20], packet[26:28])
            self[3]._set_humidity(packet[21], packet[29])
        elif self._type_id == 0x08:
            self[0]._set_temperature(packet[14:16], None, False)
            self[1]._set_rain(packet[16:18])
            self[2]._set_rain_time_span(packet[18:28])
        elif self._type_id == 0x09:
            self[0]._set_temperature(packet[14:16], packet[20:22])
            self[1]._set_temperature(packet[16:18], packet[22:24])
            self[2]._set_humidity(packet[19], packet[25])
        elif self._type_id == 0x0A:
            data: bytes = packet[14:16]
            self[0]._set_boolean(data, 0x8000)
            self[1]._set_boolean(data, 0x4000)
            self[2]._set_boolean(data, 0x2000)
            self[3]._set_boolean(data, 0x1000)
            self[4]._set_temperature(data, None, False)
        elif self._type_id == 0x0B:
            pos = 15
            for n in range(0, 5):
                self[0]._add_wind_direction(n, packet[pos])
                self[1]._add_wind_speed(n, packet[pos + 1], packet[pos], 0x01)
                self[2]._add_wind_speed(n, packet[pos + 2], packet[pos], 0x02)
                self[3]._add_wind_time_span(n, packet[pos + 3])
                pos += 4
        elif self._type_id == 0x0E:
            self[0]._set_temperature(packet[14:16], packet[19:21])
            self[0]._add_prior_temperature(packet[24:26])
            self[1]._set_humidity_hr(packet[16:18], packet[21:23], packet[26:28])
        elif self._type_id == 0x10:
            self[0]._set_boolean(packet[14:16], 0x8000)
            self[1]._set_door_window_time_span(packet[14:22])
        elif self._type_id == 0x11:
            self[2]._set_temperature(packet[14:16], packet[30:32])
            self[3]._set_humidity(packet[17], packet[33])
            self[4]._set_temperature(packet[18:20], packet[34:36])
            self[5]._set_humidity(packet[21], packet[37])
            self[6]._set_temperature(packet[22:24], packet[38:40])
            self[7]._set_humidity(packet[25], packet[41])
        elif self._type_id == 0x12:
            self[0]._set_temperature(packet[18:20], packet[25:27])
            self[1]._set_humidity(packet[20], packet[27])
            self[2]._set_humidity(packet[14], packet[21], True)
            self[3]._set_humidity(packet[15], packet[22], True)
            self[4]._set_humidity(packet[16], packet[23], True)
            self[5]._set_humidity(packet[17], packet[24], True)
        elif self._type_id == 0x15:
            self[0]._set_key_pressed(packet[14])
            self[1]._set_key_press_type(packet[14])
        elif self._type_id == 0x18:
            self[0]._set_temperature(packet[15:17], packet[20:22])
            self[1]._set_humidity(packet[17], packet[22])
            self[2]._set_air_preasure(packet[18:20], packet[23:25])
        else:
            _LOGGER.error("Unknow sensor update package %s", packet.hex().upper())
        self._last_update = packet
        _LOGGER.debug("Sensor updated: %r", self)

    @property
    def sensor_id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def model(self) -> str:
        return self._model

    @property
    def last_update(self) -> Optional[bytes]:
        return self._last_update

    @last_update.setter
    def last_update(self, value: Optional[bytes]) -> None:
        if value is not None:
            self.parse_packet(value)
        else:
            self._last_update = None

    @property
    def update_period(self) -> int:
        return self._update_period

    @property
    def parent(self) -> Any:
        return self._parent

    @property
    def counter(self) -> int:
        return self._counter

    @property
    def low_battery(self) -> bool:
        return self._low_battery

    @property
    def by_event(self) -> bool:
        return self._by_event

    @property
    def timestamp(self) -> float:
        return self._timestamp

    @property
    def measurements(self) -> List[Measurement]:
        return self._measurements
